from prosodia.core.grammar import Grammar, Language

from ..bnfrepeat import create_bnfrepeat

from ._intermediate_text import intermediate_text
from ._intermediate_transform import intermediate_transform


def create_intermediate_language() -> Language:
    return create_bnfrepeat().apply(intermediate_text)


def create_intermediate_augmentedbnf() -> Grammar[Language]:
    return Grammar(create_intermediate_language(), intermediate_transform)



import typing

from ...core import grammar as g, transform as t
from ...validation.transform_validation import annotate
from ..bnf._parser import ALLOWED_SYMBOLS

if typing.TYPE_CHECKING:
    # TypeVars should be referenced using forward references 'T' vs T so the lib
    # can do its magic
    T = typing.TypeVar('T')
    T2 = typing.TypeVar('T2')
    Addable = typing.TypeVar('Addable', str, list)


def syntax_accum(
    values: typing.Tuple[typing.Sequence[g.Rule], None]
) -> g.Language:
    rules = values[0]
    lang = g.Language.create(rules[0].name)
    eol_rule = g.Rule(
        'EOL',
        g.Syntax.create(
            g.TermGroup.create(
                g.Literal('\n')
            )
        )
    )
    eof_rule = g.Rule(
        'EOF',
        g.Syntax.create(
            g.TermGroup.create(
                g.EOFTerm()
            )
        )
    )
    lang.add_rule(eol_rule)
    lang.add_rule(eof_rule)
    for rule in rules:
        lang.add_rule(rule)
    return lang


def rule_accum(
    values: typing.Tuple[
        None,
        str,
        g.RuleName,
        str,
        None,
        str,
        None,
        typing.Sequence[g.TermGroup],
        None
    ]
) -> g.Rule:
    return g.Rule(values[2], g.Syntax(values[7]))


def expression_accum(
    values: typing.Tuple[
        g.TermGroup,
        None,
        str,
        None,
        typing.Sequence[g.TermGroup]
    ]
) -> typing.Sequence[g.TermGroup]:
    return [values[0]] + list(values[4])


def list_accum_1(
    values: typing.Tuple[g.Term]
) -> g.TermGroup:
    return g.TermGroup(list(values))


def list_accum_2(
    values: typing.Tuple[g.Term, None, g.TermGroup]
) -> g.TermGroup:
    return g.TermGroup([values[0]] + list(values[2].terms))


def text_accum_1(
    values: typing.Tuple[str]
) -> g.Literal:
    return g.Literal(values[0])


def text_accum_2(
    values: typing.Tuple[str, g.Literal]
) -> g.Literal:
    return g.Literal(values[0] + values[1].text)


def base_term_accum_rule(
    values: typing.Tuple[str, g.RuleName, str]
) -> g.Term:
    return g.RuleReference(values[1])


def literal_range_accum1(
    values: typing.Tuple[str, str, str]
) -> g.LiteralRange:
    number = int(values[1])
    return g.LiteralRange(number, number)


def literal_range_accum2(
    values: typing.Tuple[str, str, str, str, str]
) -> g.LiteralRange:
    first = int(values[1])
    second = int(values[3])
    return g.LiteralRange(first, second)


def repeat_term_accum(
    values: typing.Tuple[
        g.Term,
        str,
        typing.Tuple[int, typing.Optional[int]],
        str
    ]
) -> g.Term:
    return g.RepeatTerm(values[0], values[2][0], values[2][1])


def repeat_body_accum1(
    values: typing.Tuple[str]
) -> typing.Tuple[int, typing.Optional[int]]:
    num = int(values[0])
    return (num, num)


def repeat_body_accum2(
    values: typing.Tuple[str, str]
) -> typing.Tuple[int, typing.Optional[int]]:
    num = int(values[0])
    return (num, None)


def repeat_body_accum3(
    values: typing.Tuple[str, str, str]
) -> typing.Tuple[int, typing.Optional[int]]:
    return (int(values[0]), int(values[2]))


def list_of(
    values: typing.Tuple['T']
) -> typing.Sequence['T']:
    return [values[0]]


def push_list(
    values: typing.Tuple['T', typing.Sequence['T']]
) -> typing.Sequence['T']:
    return [values[0]] + list(values[1])


def nothing(_: typing.Tuple['T']) -> None:
    return None


def nothing2(_: typing.Tuple['T', 'T2']) -> None:
    return None


def identity(values: typing.Tuple['T']) -> 'T':
    return values[0]


def identity2(values: typing.Tuple['T']) -> 'T2':
    return values[0]  # type: ignore


def unescape(
    values: typing.Tuple[str, 'T', str]
) -> 'T':
    return values[1]


def add(
    values: typing.Tuple['Addable', 'Addable']
) -> 'Addable':
    iterable = iter(values)
    accum = next(iterable)
    for item in iterable:
        accum += item
    return accum


transform = t.LanguageTransformation.create(
    'Syntax', [syntax_accum]
)
transform <<= 'Rules', [
    annotate(list_of, T=g.Rule),
    annotate(push_list, T=g.Rule)
]
transform <<= 'Rule', [rule_accum]
transform <<= 'OptWhitespace', [
    annotate(nothing2, T=str, T2=None),
    annotate(nothing, T=str)
]
transform <<= 'Expression', [
    annotate(list_of, T=g.TermGroup),
    expression_accum
]
transform <<= 'LineEnd', [
    annotate(nothing, T=None),
    annotate(nothing2, T=None, T2=None),
]
transform <<= 'SingleLineEnd', [
    annotate(nothing2, T=None, T2=None)
]
transform <<= 'List', [
    list_accum_1,
    list_accum_2
]
transform <<= 'BaseTerm', [
    annotate(identity2, T=g.Literal, T2=g.Term),
    base_term_accum_rule,
    annotate(identity2, T=g.LiteralRange, T2=g.Term),
]
transform <<= 'Literal', [
    annotate(unescape, T=g.Literal),
    annotate(unescape, T=g.Literal),
]
transform <<= 'Text1', [
    text_accum_1,
    text_accum_2,
]
transform <<= 'Text2', [
    text_accum_1,
    text_accum_2,
]
transform <<= 'Character', [
    annotate(identity, T=str)
] * 3
transform <<= 'Letter', [
    annotate(identity, T=str)
] * 26 * 2
transform <<= 'Digit', [
    annotate(identity, T=str)
] * 2
transform <<= 'NonZeroDigit', [
    annotate(identity, T=str)
] * 9
transform <<= 'Symbol', [
    annotate(identity, T=str)
] * len(ALLOWED_SYMBOLS)
transform <<= 'Character1', [
    annotate(identity, T=str),
    annotate(identity, T=str)
]
transform <<= 'Character2', [
    annotate(identity, T=str),
    annotate(identity, T=str)
]
transform <<= 'RuleName', [
    annotate(identity, T=str),
    annotate(add, Addable=str)
]
transform <<= 'RuleEnd', [
    annotate(identity, T=str),
    annotate(add, Addable=str)
]
transform <<= 'OneRuleEnd', [
    annotate(identity, T=str),
    annotate(identity, T=str),
    annotate(add, Addable=str),
    annotate(add, Addable=str)
]
transform <<= 'EOL', [
    annotate(nothing, T=str)
]
transform <<= 'EOF', [
    annotate(nothing, T=str)
]
transform <<= 'LiteralRange', [
    literal_range_accum1,
    literal_range_accum2,
]
transform <<= 'Number', [
    annotate(identity, T=str),
    annotate(add, Addable=str)
]
transform <<= 'Digits', [
    annotate(identity, T=str),
    annotate(add, Addable=str)
]
transform <<= 'Term', [
    annotate(identity, T=g.Term),
    repeat_term_accum
]
transform <<= 'RepeatBody', [
    repeat_body_accum1,
    repeat_body_accum2,
    repeat_body_accum3,
]



from prosodia.core.grammar import Grammar, Language

from ._parser import create_language
from ._transform import transform


def create_bnf() -> Grammar[Language]:
    return Grammar(create_language(), transform)



from prosodia.core.grammar import Grammar, Language

from .. import create_bnfrange
from ._text import example_text
from ._transform import transform


def create_example_bnfrange() -> Grammar[Language]:
    return Grammar(
        create_bnfrange().apply(example_text),
        transform
    )



import typing


class Validity(object):
    def __init__(self, messages: typing.Sequence[str]) -> None:
        self.messages = messages

    @classmethod
    def valid(cls) -> 'Validity':
        return cls(list())

    @classmethod
    def invalid(cls, msg: str, *msgs: str) -> 'Validity':
        return cls([msg] + list(msgs))

    def __str__(self) -> str:
        if self:
            return 'Validity.Valid'
        else:
            return 'Validity.Invalid<{0}>'.format(len(self.messages))

    def __bool__(self) -> bool:
        return not bool(self.messages)

    def __add__(self, other: typing.Any) -> 'Validity':
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return Validity(
                list(self.messages) + list(other.messages)
            )



# pylint: disable=line-too-long,too-many-return-statements,too-many-branches,too-many-locals
import abc
from typing import Generic, TypeVar

from . import group_types as gt

OutputType = TypeVar('OutputType')


class Switch(Generic[gt.T0, OutputType], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    def __call__(self, group: gt.Group[gt.T0]) -> OutputType:
        (a,) = group
        return self.case0(a)


class Switch2(
    Generic[gt.T0, gt.T1, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    def __call__(self, group: gt.Group2) -> OutputType:
        a, b = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        else:
            raise ValueError


class Switch3(
    Generic[gt.T0, gt.T1, gt.T2, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    def __call__(self, group: gt.Group3) -> OutputType:
        a, b, c = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        else:
            raise ValueError


class Switch4(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    def __call__(self, group: gt.Group4) -> OutputType:
        a, b, c, d = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        else:
            raise ValueError


class Switch5(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    def __call__(self, group: gt.Group5) -> OutputType:
        a, b, c, d, e = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        else:
            raise ValueError


class Switch6(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    def __call__(self, group: gt.Group6) -> OutputType:
        a, b, c, d, e, f = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        else:
            raise ValueError


class Switch7(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    def __call__(self, group: gt.Group7) -> OutputType:
        a, b, c, d, e, f, g = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        else:
            raise ValueError


class Switch8(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    def __call__(self, group: gt.Group8) -> OutputType:
        a, b, c, d, e, f, g, h = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        else:
            raise ValueError


class Switch9(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    def __call__(self, group: gt.Group9) -> OutputType:
        a, b, c, d, e, f, g, h, i = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        else:
            raise ValueError


class Switch10(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, gt.T9, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    @abc.abstractmethod
    def case9(self, val: gt.T9) -> OutputType:
        pass

    def __call__(self, group: gt.Group10) -> OutputType:
        a, b, c, d, e, f, g, h, i, j = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        elif not isinstance(j, gt.NoValue):
            return self.case9(j)
        else:
            raise ValueError


class Switch11(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, gt.T9, gt.T10, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    @abc.abstractmethod
    def case9(self, val: gt.T9) -> OutputType:
        pass

    @abc.abstractmethod
    def case10(self, val: gt.T10) -> OutputType:
        pass

    def __call__(self, group: gt.Group11) -> OutputType:
        a, b, c, d, e, f, g, h, i, j, k = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        elif not isinstance(j, gt.NoValue):
            return self.case9(j)
        elif not isinstance(k, gt.NoValue):
            return self.case10(k)
        else:
            raise ValueError


class Switch12(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, gt.T9, gt.T10, gt.T11, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    @abc.abstractmethod
    def case9(self, val: gt.T9) -> OutputType:
        pass

    @abc.abstractmethod
    def case10(self, val: gt.T10) -> OutputType:
        pass

    @abc.abstractmethod
    def case11(self, val: gt.T11) -> OutputType:
        pass

    def __call__(self, group: gt.Group12) -> OutputType:
        a, b, c, d, e, f, g, h, i, j, k, l = group  # noqa: E741
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        elif not isinstance(j, gt.NoValue):
            return self.case9(j)
        elif not isinstance(k, gt.NoValue):
            return self.case10(k)
        elif not isinstance(l, gt.NoValue):
            return self.case11(l)
        else:
            raise ValueError


class Switch13(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, gt.T9, gt.T10, gt.T11, gt.T12, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    @abc.abstractmethod
    def case9(self, val: gt.T9) -> OutputType:
        pass

    @abc.abstractmethod
    def case10(self, val: gt.T10) -> OutputType:
        pass

    @abc.abstractmethod
    def case11(self, val: gt.T11) -> OutputType:
        pass

    @abc.abstractmethod
    def case12(self, val: gt.T12) -> OutputType:
        pass

    def __call__(self, group: gt.Group13) -> OutputType:
        a, b, c, d, e, f, g, h, i, j, k, l, m = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        elif not isinstance(j, gt.NoValue):
            return self.case9(j)
        elif not isinstance(k, gt.NoValue):
            return self.case10(k)
        elif not isinstance(l, gt.NoValue):
            return self.case11(l)
        elif not isinstance(m, gt.NoValue):
            return self.case12(m)
        else:
            raise ValueError


class Switch14(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, gt.T9, gt.T10, gt.T11, gt.T12, gt.T13, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    @abc.abstractmethod
    def case9(self, val: gt.T9) -> OutputType:
        pass

    @abc.abstractmethod
    def case10(self, val: gt.T10) -> OutputType:
        pass

    @abc.abstractmethod
    def case11(self, val: gt.T11) -> OutputType:
        pass

    @abc.abstractmethod
    def case12(self, val: gt.T12) -> OutputType:
        pass

    @abc.abstractmethod
    def case13(self, val: gt.T13) -> OutputType:
        pass

    def __call__(self, group: gt.Group14) -> OutputType:
        a, b, c, d, e, f, g, h, i, j, k, l, m, n = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        elif not isinstance(j, gt.NoValue):
            return self.case9(j)
        elif not isinstance(k, gt.NoValue):
            return self.case10(k)
        elif not isinstance(l, gt.NoValue):
            return self.case11(l)
        elif not isinstance(m, gt.NoValue):
            return self.case12(m)
        elif not isinstance(n, gt.NoValue):
            return self.case13(n)
        else:
            raise ValueError


class Switch15(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, gt.T9, gt.T10, gt.T11, gt.T12, gt.T13, gt.T14, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    @abc.abstractmethod
    def case9(self, val: gt.T9) -> OutputType:
        pass

    @abc.abstractmethod
    def case10(self, val: gt.T10) -> OutputType:
        pass

    @abc.abstractmethod
    def case11(self, val: gt.T11) -> OutputType:
        pass

    @abc.abstractmethod
    def case12(self, val: gt.T12) -> OutputType:
        pass

    @abc.abstractmethod
    def case13(self, val: gt.T13) -> OutputType:
        pass

    @abc.abstractmethod
    def case14(self, val: gt.T14) -> OutputType:
        pass

    def __call__(self, group: gt.Group15) -> OutputType:
        a, b, c, d, e, f, g, h, i, j, k, l, m, n, o = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        elif not isinstance(j, gt.NoValue):
            return self.case9(j)
        elif not isinstance(k, gt.NoValue):
            return self.case10(k)
        elif not isinstance(l, gt.NoValue):
            return self.case11(l)
        elif not isinstance(m, gt.NoValue):
            return self.case12(m)
        elif not isinstance(n, gt.NoValue):
            return self.case13(n)
        elif not isinstance(o, gt.NoValue):
            return self.case14(o)
        else:
            raise ValueError


class Switch16(
    Generic[gt.T0, gt.T1, gt.T2, gt.T3, gt.T4, gt.T5, gt.T6, gt.T7, gt.T8, gt.T9, gt.T10, gt.T11, gt.T12, gt.T13, gt.T14, gt.T15, OutputType],
    metaclass=abc.ABCMeta,
):
    @abc.abstractmethod
    def case0(self, val: gt.T0) -> OutputType:
        pass

    @abc.abstractmethod
    def case1(self, val: gt.T1) -> OutputType:
        pass

    @abc.abstractmethod
    def case2(self, val: gt.T2) -> OutputType:
        pass

    @abc.abstractmethod
    def case3(self, val: gt.T3) -> OutputType:
        pass

    @abc.abstractmethod
    def case4(self, val: gt.T4) -> OutputType:
        pass

    @abc.abstractmethod
    def case5(self, val: gt.T5) -> OutputType:
        pass

    @abc.abstractmethod
    def case6(self, val: gt.T6) -> OutputType:
        pass

    @abc.abstractmethod
    def case7(self, val: gt.T7) -> OutputType:
        pass

    @abc.abstractmethod
    def case8(self, val: gt.T8) -> OutputType:
        pass

    @abc.abstractmethod
    def case9(self, val: gt.T9) -> OutputType:
        pass

    @abc.abstractmethod
    def case10(self, val: gt.T10) -> OutputType:
        pass

    @abc.abstractmethod
    def case11(self, val: gt.T11) -> OutputType:
        pass

    @abc.abstractmethod
    def case12(self, val: gt.T12) -> OutputType:
        pass

    @abc.abstractmethod
    def case13(self, val: gt.T13) -> OutputType:
        pass

    @abc.abstractmethod
    def case14(self, val: gt.T14) -> OutputType:
        pass

    @abc.abstractmethod
    def case15(self, val: gt.T15) -> OutputType:
        pass

    def __call__(self, group: gt.Group16) -> OutputType:
        a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p = group
        if not isinstance(a, gt.NoValue):
            return self.case0(a)
        elif not isinstance(b, gt.NoValue):
            return self.case1(b)
        elif not isinstance(c, gt.NoValue):
            return self.case2(c)
        elif not isinstance(d, gt.NoValue):
            return self.case3(d)
        elif not isinstance(e, gt.NoValue):
            return self.case4(e)
        elif not isinstance(f, gt.NoValue):
            return self.case5(f)
        elif not isinstance(g, gt.NoValue):
            return self.case6(g)
        elif not isinstance(h, gt.NoValue):
            return self.case7(h)
        elif not isinstance(i, gt.NoValue):
            return self.case8(i)
        elif not isinstance(j, gt.NoValue):
            return self.case9(j)
        elif not isinstance(k, gt.NoValue):
            return self.case10(k)
        elif not isinstance(l, gt.NoValue):
            return self.case11(l)
        elif not isinstance(m, gt.NoValue):
            return self.case12(m)
        elif not isinstance(n, gt.NoValue):
            return self.case13(n)
        elif not isinstance(o, gt.NoValue):
            return self.case14(o)
        elif not isinstance(p, gt.NoValue):
            return self.case15(p)
        else:
            raise ValueError



from itertools import chain

from prosodia.core.grammar import Language
from ...core import grammar as g

ALLOWED_SYMBOLS = r'| !#$%&()*+,-./:;>=<?@[\]^_`{}~'


def create_language() -> Language:
    lang = g.Language.create('Syntax')
    lang.add_rule(
        g.Rule(
            'Syntax',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Rules'),
                    g.RuleReference('EOF')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Rules',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Rule')
                ),
                g.TermGroup.create(
                    g.RuleReference('Rule'),
                    g.RuleReference('Rules')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Rule',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('OptWhitespace'),
                    g.Literal('<'),
                    g.RuleReference('RuleName'),
                    g.Literal('>'),
                    g.RuleReference('OptWhitespace'),
                    g.Literal('::='),
                    g.RuleReference('OptWhitespace'),
                    g.RuleReference('Expression'),
                    g.RuleReference('LineEnd')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'OptWhitespace',
            g.Syntax.create(
                g.TermGroup.create(
                    g.Literal(' '),
                    g.RuleReference('OptWhitespace')
                ),
                g.TermGroup.create(
                    g.Literal('')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Expression',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('List')
                ),
                g.TermGroup.create(
                    g.RuleReference('List'),
                    g.RuleReference('OptWhitespace'),
                    g.Literal('|'),
                    g.RuleReference('OptWhitespace'),
                    g.RuleReference('Expression')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'LineEnd',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('SingleLineEnd'),
                ),
                g.TermGroup.create(
                    g.RuleReference('SingleLineEnd'),
                    g.RuleReference('LineEnd'),
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'SingleLineEnd',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('OptWhitespace'),
                    g.RuleReference('EOL')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'List',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Term')
                ),
                g.TermGroup.create(
                    g.RuleReference('Term'),
                    g.RuleReference('OptWhitespace'),
                    g.RuleReference('List')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Term',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Literal')
                ),
                g.TermGroup.create(
                    g.Literal('<'),
                    g.RuleReference('RuleName'),
                    g.Literal('>'),
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Literal',
            g.Syntax.create(
                g.TermGroup.create(
                    g.Literal('"'),
                    g.RuleReference('Text1'),
                    g.Literal('"'),
                ),
                g.TermGroup.create(
                    g.Literal("'"),
                    g.RuleReference('Text2'),
                    g.Literal("'"),
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Text1',
            g.Syntax.create(
                g.TermGroup.create(
                    g.Literal('')
                ),
                g.TermGroup.create(
                    g.RuleReference('Character1'),
                    g.RuleReference('Text1'),
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Text2',
            g.Syntax.create(
                g.TermGroup.create(
                    g.Literal('')
                ),
                g.TermGroup.create(
                    g.RuleReference('Character2'),
                    g.RuleReference('Text2'),
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Character',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Letter')
                ),
                g.TermGroup.create(
                    g.RuleReference('Digit')
                ),
                g.TermGroup.create(
                    g.RuleReference('Symbol')
                ),
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Letter',
            g.Syntax.create(
                *(
                    g.TermGroup.create(
                        g.Literal(letter)
                    ) for letter in chain(
                        (chr(ord('A') + offset) for offset in range(26)),
                        (chr(ord('a') + offset) for offset in range(26))
                    )
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Digit',
            g.Syntax.create(
                *(
                    g.TermGroup.create(
                        g.Literal(chr(ord('0') + offset))
                    ) for offset in range(10)
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Symbol',
            g.Syntax.create(
                *(
                    g.TermGroup.create(
                        g.Literal(symbol)
                    ) for symbol in ALLOWED_SYMBOLS
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Character1',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Character')
                ),
                g.TermGroup.create(
                    g.Literal("'")
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'Character2',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Character')
                ),
                g.TermGroup.create(
                    g.Literal('"')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'RuleName',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Letter')
                ),
                g.TermGroup.create(
                    g.RuleReference('Letter'),
                    g.RuleReference('RuleEnd')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'RuleEnd',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('OneRuleEnd')
                ),
                g.TermGroup.create(
                    g.RuleReference('OneRuleEnd'),
                    g.RuleReference('RuleEnd')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'OneRuleEnd',
            g.Syntax.create(
                g.TermGroup.create(
                    g.RuleReference('Letter')
                ),
                g.TermGroup.create(
                    g.RuleReference('Digit')
                ),
                g.TermGroup.create(
                    g.Literal('-'),
                    g.RuleReference('Letter')
                ),
                g.TermGroup.create(
                    g.Literal('-'),
                    g.RuleReference('Digit')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'EOL',
            g.Syntax.create(
                g.TermGroup.create(
                    g.Literal('\n')
                )
            )
        )
    )
    lang.add_rule(
        g.Rule(
            'EOF',
            g.Syntax.create(
                g.TermGroup.create(
                    g.EOFTerm()
                )
            )
        )
    )
    return lang



from functools import wraps
from inspect import signature
from typing import Iterable, Callable, get_type_hints, Union, Tuple

from .validity import Validity


def annotate(func: Callable, **hints: Union[type, None]) -> Callable:
    try:
        get_type_hints(func, {}, {})
    except NameError:
        pass
    else:
        raise ValueError(
            'expecting to be able to substitute some forward references'
        )

    @wraps(func)
    def annotated(*args: object, **kwargs: object) -> object:
        return func(*args, **kwargs)
    annotated.__annotations__ = get_type_hints(func, hints, {})
    return annotated


def get_return_type(func: Callable) -> type:
    sig = signature(func, follow_wrapped=False)
    anno = sig.return_annotation
    if anno is sig.empty:
        raise TypeError('func doesn\'t have a return type annotation')
    return anno


def check_composability(
    argument_types: Iterable[type],
    target_func: Callable
) -> Validity:
    argument_types = [Tuple[tuple(argument_types)]]
    parameter_hints = get_type_hints(target_func)
    parameter_types = [
        parameter_hints[param_name]
        for param_name in signature(target_func).parameters.keys()
    ]
    if argument_types == parameter_types:
        return Validity.valid()
    else:
        return Validity.invalid(
            'arguments are not composable with parameters: ({0}), ({1})'.format(
                argument_types, parameter_types
            )
        )


def check_isomorphic(
    funcs: Iterable[Callable]
) -> bool:
    first, *rest = (get_return_type(func) for func in funcs)
    return all(
        first == rtype
        for rtype in rest
    )



from typing import Sequence, Tuple

from prosodia.core.grammar import Language
from prosodia.core.transform import LanguageTransformation
from prosodia.validation.transform_validation import annotate
from prosodia.validation.group_types import Group2, NoValue

from .._transform_helpers import add, identity

from .._intermediate_parser import create_intermediate_augmentedbnf
from .text import freebies_text


def add_freebie_rules(lang: Language) -> Language:
    abnf = create_intermediate_augmentedbnf()
    freebies = abnf.apply(freebies_text).rules
    for rule in freebies.values():
        lang.add_rule(rule)
    return lang


def add_freebie_transforms(
    lt: LanguageTransformation
) -> LanguageTransformation:
    lt <<= 'ALPHA', [annotate(identity, T=str)] * 2
    lt <<= 'BIT', [annotate(identity, T=str)] * 2
    lt <<= 'CHAR', [annotate(identity, T=str)]
    lt <<= 'CR', [annotate(identity, T=str)]
    lt <<= 'CRLF', [annotate(add, Addable=str)]
    lt <<= 'CTL', [annotate(identity, T=str)] * 2
    lt <<= 'DIGIT', [annotate(identity, T=str)]
    lt <<= 'DQUOTE', [annotate(identity, T=str)]
    lt <<= 'HEXDIG', [annotate(identity, T=str)] * 7
    lt <<= 'HTAB', [annotate(identity, T=str)]
    lt <<= 'LF', [annotate(identity, T=str)]
    lt <<= 'LWSP', [lwsp_accum]
    lt <<= 'OCTET', [annotate(identity, T=str)]
    lt <<= 'SP', [annotate(identity, T=str)]
    lt <<= 'VCHAR', [annotate(identity, T=str)]
    lt <<= 'WSP', [annotate(identity, T=str)] * 2
    return lt


def lwsp_accum(
    values: Tuple[Sequence[
        Group2[
            Tuple[str],
            Tuple[str, str]
        ]
    ]]
) -> str:
    total = ''
    for i in values[0]:
        cand1, cand2 = i
        if not isinstance(cand1, NoValue):
            total += cand1[0]
        elif not isinstance(cand2, NoValue):
            total += cand2[0] + cand2[1]
    return total



import typing

from .grammar import RuleName
from . import grammar as g
from .resolvable import (
    Resolvable,
    ResolvableFunc,
    ResolvablePair,
    resolve,
    resolve_map
)
from .tree import Node, RuleNode
from ..validation.validity import Validity
from ..validation.transform_validation import (
    check_composability, check_isomorphic)

InputType = typing.TypeVar('InputType')
OutputType = typing.TypeVar('OutputType')
T = typing.TypeVar('T')


class LanguageTransformation(typing.Generic[T]):
    def __init__(
        self,
        transformation_rules: typing.Dict[
            RuleName, 'RuleTransformation[typing.Any]'
        ],
        unused_root_rule: 'RuleTransformation[T]'
    ) -> None:
        self.transformation_rules = transformation_rules

    @classmethod
    def create(
        cls,
        rule_name: RuleName,
        root_accums: typing.Sequence[typing.Callable[..., T]]
    ) -> 'LanguageTransformation[T]':
        rt: RuleTransformation[T] = RuleTransformation(
            rule_name,
            SyntaxTransformation(
                tuple(
                    TermGroupTransformation(t) for t in typing.cast(
                        typing.Sequence[typing.Callable[..., Resolvable[T]]],
                        root_accums
                    )
                )
            )
        )
        return cls({rule_name: rt}, rt)

    def add_rule_transformation(
        self,
        rt: 'RuleTransformation'
    ) -> 'LanguageTransformation':
        self.transformation_rules[rt.rule_name] = rt
        return self

    def transform(self, node: Node) -> typing.Any:
        return resolve(node.transform(self))

    def __ilshift__(
        self,
        info: typing.Tuple[
            RuleName,
            typing.Sequence[typing.Callable[..., typing.Any]]
        ]
    ) -> 'LanguageTransformation':
        rule_name, transforms = info
        rt = RuleTransformation(
            rule_name,
            SyntaxTransformation(
                tuple(
                    TermGroupTransformation(t) for t in transforms
                )
            )
        )
        self.add_rule_transformation(rt)
        return self

    def validate(self, lang: g.Language) -> Validity:
        if self.transformation_rules.keys() != lang.rules.keys():
            return Validity.invalid(
                'lang does not have the same set of rule names. transform '
                'extras {0}, lang extras {1}.'.format(
                    self.transformation_rules.keys() - lang.rules.keys(),
                    lang.rules.keys() - self.transformation_rules.keys(),
                )
            )
        else:
            return sum(
                (
                    rule_t.validate(lang.rules[k], self)
                    for k, rule_t in self.transformation_rules.items()
                ),
                Validity.valid()
            )


class RuleTransformation(typing.Generic[OutputType]):
    def __init__(
        self,
        rule_name: RuleName,
        tf_syntax: 'SyntaxTransformation[OutputType]'
    ) -> None:
        self.rule_name = rule_name
        self.tf_syntax = tf_syntax

    def transform(
        self,
        rule_node: RuleNode,
        lang: 'LanguageTransformation'
    ) -> Resolvable[OutputType]:
        return ResolvableFunc(
            self.tf_syntax.transform,
            rule_node.children,
            rule_node.term_group_id,
            lang
        )

    def validate(
        self,
        rule: g.Rule,
        lt: 'LanguageTransformation'
    ) -> Validity:
        if self.rule_name != rule.name:
            return Validity.invalid(
                'rule transform is named differently than the rule'
            )
        else:
            validity = self.tf_syntax.validate(rule.syntax, lt)
            if not validity:
                v = validity + Validity.invalid(
                    '{0} rule is not valid'.format(repr(self.rule_name))
                )
                return v
            else:
                return Validity.valid()


class SyntaxTransformation(typing.Generic[OutputType]):
    def __init__(
        self,
        tf_term_groups: typing.Sequence['TermGroupTransformation[OutputType]']
    ) -> None:
        self.tf_term_groups = tf_term_groups

    @classmethod
    def create(
        cls,
        *tf_term_groups: 'TermGroupTransformation[OutputType]'
    ) -> 'SyntaxTransformation':
        return cls(tf_term_groups)

    def transform(
        self,
        values: typing.Sequence[Node],
        index: int,
        lang: 'LanguageTransformation'
    ) -> Resolvable[OutputType]:
        return ResolvableFunc(
            self.tf_term_groups[index].transform,
            values,
            lang
        )

    def validate(
        self,
        syntax: g.Syntax,
        lt: 'LanguageTransformation'
    ) -> Validity:
        if len(self.tf_term_groups) != len(syntax.term_groups):
            return Validity.invalid(
                'the syntax has a different number of term groups than the '
                'transformation. transform group count {0}, language group '
                'count {1}'.format(
                    len(self.tf_term_groups), len(syntax.term_groups)
                )
            )
        elif not check_isomorphic(
            tgt.accumulator for tgt in self.tf_term_groups
        ):
            return Validity.invalid(
                'not all term group transformations return the same type'
            )
        else:
            return sum(
                (
                    tgt.validate(tg, lt)
                    for tgt, tg in zip(self.tf_term_groups, syntax.term_groups)
                ),
                Validity.valid()
            )


class TermGroupTransformation(typing.Generic[OutputType]):
    def __init__(
        self,
        accumulator: typing.Callable[[typing.Any], Resolvable[OutputType]]
    ) -> None:
        self.accumulator = accumulator

    def transform(
        self,
        values: typing.Sequence[Node],
        lang: 'LanguageTransformation'
    ) -> Resolvable[OutputType]:
        return ResolvablePair(
            resolve_map(
                values,
                lambda v: v.transform(lang)
            ),
            self.accumulator
        )

    def validate(
        self,
        term_group: g.TermGroup,
        lt: 'LanguageTransformation'
    ) -> Validity:
        transform_types = [
            term.get_transform_type(lt)
            for term in term_group.terms
        ]
        is_composable = check_composability(transform_types, self.accumulator)
        if not is_composable:
            return is_composable + Validity.invalid(
                'rule references in the term group are not composable: '
                + ' '.join(repr(t) for t in term_group.terms)
            )
        else:
            return Validity.valid()


class LazySequenceTransform(typing.Sequence[typing.Any]):
    def __init__(
        self,
        initial_values: typing.Sequence[Node],
        lang: 'LanguageTransformation',
        cache: typing.Dict[int, typing.Any]
    ) -> None:
        super().__init__()
        self.initial_values = initial_values
        self.lang = lang
        self.cache = cache

    @classmethod
    def create(
        cls,
        nodes: typing.Sequence[Node],
        lang: 'LanguageTransformation'
    ) -> 'LazySequenceTransform':
        return cls(nodes, lang, dict())

    @typing.overload
    def __getitem__(self, i: int) -> typing.Any:
        pass

    @typing.overload
    def __getitem__(self, s: slice) -> typing.Sequence:  # pylint: disable=function-redefined
        pass

    def __getitem__(self, x: typing.Union[int, slice]) -> typing.Any:\
            # pylint: disable=function-redefined
        if isinstance(x, int):
            if x not in self.cache:
                self.cache[x] = self.initial_values[x].transform(self.lang)
            return self.cache[x]
        elif isinstance(x, slice):
            return tuple(
                self[i]
                for i in range(len(self))[x]
            )
        else:
            raise TypeError(
                'Expected `x` to be a Union[int, slice] but got {0}'.format(
                    type(x)
                )
            )

    def __len__(self) -> int:
        return len(self.initial_values)

    def __iter__(self) -> typing.Iterator:
        return (self[i] for i in range(len(self)))



import abc
from functools import partial
import typing

from .tree import Node, LiteralNode, RuleNode, MultiNode
from ..validation.validity import Validity
from ..validation import group_types as gt
from ..validation.transform_validation import get_return_type
if typing.TYPE_CHECKING:
    from .transform import LanguageTransformation  # pylint: disable=unused-import

RuleName = str
MatchResult = typing.Tuple['_SmartText', Node]
RuleReferenceCache = typing.MutableMapping[
    typing.Tuple[int, '_SmartText', int],
    typing.Sequence[MatchResult]
]
T = typing.TypeVar('T')


class Grammar(typing.Generic[T]):
    def __init__(
        self,
        language: 'Language',
        transform: 'LanguageTransformation[T]',
        allow_partial_matches: bool = True
    ) -> None:
        self.language = language
        self.transform = transform
        self.allow_partial_matches = allow_partial_matches

    def apply(self, text: str) -> T:
        return self.transform.transform(
            self.language.parse(
                text,
                self.allow_partial_matches
            )
        )

    def validate(self) -> Validity:
        return self.language.validate() + self.transform.validate(self.language)


class _SmartText(object):
    __slots__ = ['_raw_text', '_start', '_end']

    def __init__(
        self,
        raw_text: str,
        *,
        _start: int = None,
        _end: int = None
    ) -> None:
        self._raw_text = raw_text
        self._start = 0 if _start is None else _start
        self._end = len(raw_text) if _end is None else _end

    def __len__(self) -> int:
        return self._end - self._start

    def log_front(self, size: int, *args: object) -> None:
        segment = self._raw_text[
            self._start:self._end if len(self) < size else self._start + size
        ]
        print(len(self), self._start, self._end, repr(segment), *args)

    def startswith(self, target: str, case_sensitive: bool = True) -> bool:
        src = self._raw_text[self._start: self._start + len(target)]
        if case_sensitive:
            return src == target
        else:
            return src.lower() == target.lower()

    @typing.overload
    def __getitem__(self, index: int) -> str:
        pass

    @typing.overload
    def __getitem__(self, index: slice) -> '_SmartText':  # pylint: disable=function-redefined
        pass

    def __getitem__(self, index: typing.Any) -> typing.Any:  # pylint: disable=function-redefined
        if isinstance(index, int):
            return self._raw_text[self._start + index]
        elif isinstance(index, slice):
            if index.step is not None and index.step != 1:
                raise TypeError('cannot use step of not 1 with _SmartText')
            start = self._start + (index.start or 0)
            if index.stop is None:
                end = self._end
            else:
                end = start + index.stop
            return _SmartText(self._raw_text, _start=start, _end=end)
        else:
            raise TypeError('expecting an int or a slice')

    def _as_tuple(self) -> typing.Tuple[object, ...]:
        return (self._raw_text, self._start, self._end)

    def __hash__(self) -> int:
        return hash(self._as_tuple())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _SmartText):
            return NotImplemented
        return (
            type(self) is type(other)
            and self._as_tuple() == other._as_tuple()
        )


class NoMatches(Exception):
    pass


class TooManyMatches(Exception):
    def __init__(self, matches: typing.Sequence[Node]) -> None:
        super().__init__(self)
        self.matches = matches


class Language(object):
    """Collection of rules"""
    def __init__(
        self,
        rules: typing.Dict[RuleName, 'Rule'],
        root_rule: RuleName,
        debug: bool = False
    ) -> None:
        self.rules = rules
        self.root_rule = root_rule
        self.debug = debug

    @classmethod
    def create(cls, root_rule: RuleName) -> 'Language':
        return cls(dict(), root_rule)

    def log_match(self, text: _SmartText, *args: object) -> None:
        if self.debug:
            text.log_front(20, *args)

    def add_rule(self, rule: 'Rule') -> 'Language':
        if rule.name in self.rules:
            raise ValueError(
                'Rule with that name is already in this Language: {0}'.format(
                    rule.name
                )
            )
        self.rules[rule.name] = rule
        return self

    def get_rule(self, rule_name: RuleName) -> 'Rule':
        return self.rules[rule_name]

    def add_to_rule(
        self,
        rule_name: RuleName,
        tgs: typing.Sequence['TermGroup']
    ) -> None:
        self.rules[rule_name].syntax.term_groups = (
            list(self.rules[rule_name].syntax.term_groups) + list(tgs)
        )

    def parse(self, raw_text: str, allow_partial_matches: bool = True) -> Node:
        text = _SmartText(raw_text)
        matches = tuple(self._parse_all(text, allow_partial_matches, {}))
        if not matches:
            raise NoMatches
        elif len(matches) > 1:
            for m in matches:
                print(m.draw())
            raise TooManyMatches(matches)
        else:
            return matches[0]

    def _parse_all(
        self,
        text: _SmartText,
        allow_partial_matches: bool,
        cache: RuleReferenceCache,
    ) -> typing.Iterable[Node]:
        root = self.get_rule(self.root_rule)
        matches = root.match(text, self, cache)
        for leftover, node in matches:
            if allow_partial_matches or not leftover:
                yield node

    def equals(self, other: 'Language') -> Validity:
        if self.root_rule != other.root_rule:
            return Validity.invalid('Langage: root rule is different')
        elif self.rules.keys() != other.rules.keys():
            return Validity.invalid('Language: set of rule names are different')

        validity = sum(
            (
                r.equals(other.rules[k])
                for k, r in self.rules.items()
            ),
            Validity.valid()
        )
        if not validity:
            return validity + Validity.invalid('Language: rules are different')
        else:
            return Validity.valid()

    def validate(self) -> Validity:
        if self.root_rule not in self.rules:
            return Validity.invalid('root rule is not present in language')
        else:
            return sum(
                (
                    rule.validate(self)
                    for rule in self.rules.values()
                ),
                Validity.valid()
            )


class Rule(object):
    """Syntax rule

    each transformation in the map must match the format of the syntax
    """
    def __init__(
        self,
        name: RuleName,
        syntax: 'Syntax'
    ) -> None:
        self.name = name
        self.syntax = syntax

    def match(
        self,
        text: _SmartText,
        lang: 'Language',
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        return self.syntax.match(text, self.name, lang, cache)

    def equals(self, other: 'Rule') -> Validity:
        if self.name != other.name:
            return Validity.invalid('Rule: rule names are different')

        validity = self.syntax.equals(other.syntax)
        if not validity:
            return validity + Validity.invalid('Rule: syntaxes are different')
        else:
            return Validity.valid()

    def validate(self, lang: 'Language') -> Validity:
        return self.syntax.validate(self.name, lang)


class Syntax(object):
    """Collection of Term lists

    a syntax matches if one of its term lists matches the plaintext
    """
    def __init__(self, term_groups: typing.Sequence['TermGroup']) -> None:
        self.term_groups = term_groups

    @classmethod
    def create(cls, *term_groups: 'TermGroup') -> 'Syntax':
        return cls(term_groups)

    def match(
        self,
        text: _SmartText,
        rule_name: 'RuleName',
        lang: 'Language',
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        for index, term_list in enumerate(self.term_groups):
            for leftover, terms in term_list.match(text, lang, cache):
                node = RuleNode(rule_name, index, terms)
                yield leftover, node

    def equals(self, other: 'Syntax') -> Validity:
        if len(self.term_groups) != len(other.term_groups):
            return Validity.invalid(
                'Syntax: number of term groups are different'
            )
        validity = sum(
            (
                i.equals(j)
                for i, j in zip(self.term_groups, other.term_groups)
            ),
            Validity.valid()
        )
        if not validity:
            return validity + Validity.invalid(
                'Syntax: term groups are different'
            )
        else:
            return Validity.valid()

    def validate(self, rule_name: RuleName, lang: 'Language') -> Validity:
        dupes = tuple(
            (i, j, x)
            for i, x in enumerate(self.term_groups)
            for j, y in enumerate(self.term_groups)
            if i != j and x.equals(y)
        )
        if dupes:
            return Validity.invalid('syntax has duplicate term groups')
        else:
            return sum(
                (tg.validate(rule_name, lang) for tg in self.term_groups),
                Validity.valid()
            )


class TermGroup(object):
    """Collection of Terms"""
    def __init__(self, terms: typing.Sequence['Term']) -> None:
        self.terms = terms

    @classmethod
    def create(cls, *terms: 'Term') -> 'TermGroup':
        return cls(terms)

    def _match_impl(
        self,
        text: _SmartText,
        term_index: int,
        lang: 'Language',
        cache: RuleReferenceCache,
    ) -> typing.Iterable[typing.Tuple[_SmartText, typing.Sequence[Node]]]:
        term = self.terms[term_index]
        next_term_index = term_index + 1
        is_last_term = next_term_index >= len(self.terms)
        for leftover, match in term.match(text, lang, cache):
            if is_last_term:
                yield leftover, (match,)
            else:
                for final_leftover, child_matches in self._match_impl(
                    leftover,
                    next_term_index,
                    lang,
                    cache,
                ):
                    yield final_leftover, (match,) + tuple(child_matches)

    def match(
        self,
        text: _SmartText,
        lang: 'Language',
        cache: RuleReferenceCache,
    ) -> typing.Iterable[typing.Tuple[_SmartText, typing.Sequence[Node]]]:
        return self._match_impl(text, 0, lang, cache)

    def equals(self, other: 'TermGroup') -> Validity:
        if len(self.terms) != len(other.terms):
            return Validity.invalid(
                'TermGroup: number of terms are different'
            )
        validity = sum(
            (
                i.equals(j)
                for i, j in zip(self.terms, other.terms)
            ),
            Validity.valid()
        )
        if not validity:
            return validity + Validity.invalid(
                'TermGroup: terms are different'
            )
        else:
            return Validity.valid()

    def validate(self, rule_name: RuleName, lang: 'Language') -> Validity:
        if not self.terms:
            return Validity.invalid(
                'term group needs at least one term'
            )
        elif (
            len(self.terms) > 1
            and any(t.equals(Literal('')) for t in self.terms)
        ):
            return Validity.invalid(
                'shouldnt have an empty literal in a term group with more than '
                'one term'
            )
        elif self.terms[0].equals(RuleReference(rule_name)):
            return Validity.invalid(
                'term group cannot start with its own rule'
            )
        else:
            return sum(
                (t.validate(lang) for t in self.terms),
                Validity.valid()
            )


class Term(object, metaclass=abc.ABCMeta):
    """Unit of a syntax"""
    @abc.abstractmethod
    def match(
        self,
        text: _SmartText,
        lang: 'Language',
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        raise NotImplementedError

    @abc.abstractmethod
    def validate(self, lang: 'Language') -> Validity:
        raise NotImplementedError

    @abc.abstractmethod
    def equals(self, other: 'Term') -> Validity:
        raise NotImplementedError

    @abc.abstractmethod
    def get_transform_type(self, lt: 'LanguageTransformation') -> type:
        raise NotImplementedError


class RuleReference(Term):
    """Term that represents a nested rule"""
    def __init__(self, rule_name: RuleName) -> None:
        self.rule_name = rule_name

    def match(
        self,
        text: _SmartText,
        lang: 'Language',
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        key = id(self), text, id(lang)
        if key in cache:
            yield from iter(cache[key])
            return
        results: typing.List[MatchResult] = []
        cache[key] = results
        matches = lang.get_rule(self.rule_name).match(text, lang, cache)
        actual_match = False
        for match in matches:
            results.append(match)
            lang.log_match(match[0], 'rule ref', self.rule_name)
            if not actual_match:
                if match[0] != text:
                    # Since matching only moves forwards, as long as we match
                    # any part of the text then we know that we will finish
                    # exhausting this iterable before we hit this cache entry
                    actual_match = True
                    yield from results
            else:
                yield match
        if not actual_match and results:
            yield from results

    def equals(self, other: Term) -> Validity:
        if not isinstance(other, RuleReference):
            return Validity.invalid(
                'RuleReference: other is not rule reference'
            )
        elif self.rule_name != other.rule_name:
            return Validity.invalid('RuleReference: different rule names')
        else:
            return Validity.valid()

    def __repr__(self) -> str:
        return '<RuleReference(Term) {0}>'.format(self.rule_name)

    def validate(self, lang: 'Language') -> Validity:
        if self.rule_name not in lang.rules:
            return Validity.invalid(
                'rule reference points to a rule that does not exist: {0!r}'
                .format(self.rule_name)
            )
        else:
            return Validity.valid()

    def get_transform_type(self, lt: 'LanguageTransformation') -> type:
        rules = lt.transformation_rules
        return get_return_type(
            rules[self.rule_name].tf_syntax.tf_term_groups[0].accumulator
        )


class Literal(Term):
    """Term that represents a plaintext literal"""
    def __init__(self, text: str, case_sensitive: bool = True) -> None:
        self.text = text
        self.case_sensitive = case_sensitive

    def match(
        self,
        text: _SmartText,
        lang: Language,
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        if text.startswith(self.text, self.case_sensitive):
            lang.log_match(text, 'literal', self.text)
            yield text[len(self.text):], LiteralNode(self.text)

    def __repr__(self) -> str:
        return '<Literal(Term) {0!r}>'.format(self.text)

    def equals(self, other: Term) -> Validity:
        if not isinstance(other, Literal):
            return Validity.invalid('Literal: other is not a literal')
        elif self.text != other.text:
            return Validity.invalid('Literal: text is different')
        else:
            return Validity.valid()

    def validate(self, lang: Language) -> Validity:
        return Validity.valid()

    def get_transform_type(self, lt: 'LanguageTransformation') -> type:
        return get_return_type(LiteralNode.transform)


class LiteralRange(Term):
    def __init__(self, min_value: int, max_value: int) -> None:
        self.min_value = min_value
        self.max_value = max_value

    def match(
        self,
        text: _SmartText,
        lang: Language,
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        if text:
            first_char = text[0]
            if self.min_value <= ord(first_char) <= self.max_value:
                lang.log_match(text, 'range', self.min_value, self.max_value)
                yield text[1:], LiteralNode(first_char)

    def __repr__(self) -> str:
        return '<LiteralRange(Term) {0}, {1}>'.format(
            self.min_value,
            self.max_value
        )

    def equals(self, other: Term) -> Validity:
        if not isinstance(other, LiteralRange):
            return Validity.invalid(
                'LiteralRange: other is not a literal range'
            )
        elif (
            self.min_value != other.min_value
            or self.max_value != other.max_value
        ):
            return Validity.invalid('LiteralRange: range values are different')
        else:
            return Validity.valid()

    def validate(self, lang: Language) -> Validity:
        if self.max_value < self.min_value:
            return Validity.invalid('LiteralRange: max cannot be less than min')
        elif self.max_value < 0 or self.min_value < 0:
            return Validity.invalid(
                'LiteralRange: cannot have negative range values'
            )
        else:
            return Validity.valid()

    def get_transform_type(self, lt: 'LanguageTransformation') -> type:
        return get_return_type(LiteralNode.transform)


class EOFTerm(Term):
    def match(
        self,
        text: _SmartText,
        lang: 'Language',
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        if not text:
            lang.log_match(text, 'EOF')
            yield text, LiteralNode('')

    def __repr__(self) -> str:
        return '<EOFTerm(Term)>'

    def equals(self, other: Term) -> Validity:
        if not isinstance(other, EOFTerm):
            return Validity.invalid('EOFTerm: other is not an EOFTerm')
        else:
            return Validity.valid()

    def validate(self, lang: Language) -> Validity:
        return Validity.valid()

    def get_transform_type(self, lt: 'LanguageTransformation') -> type:
        return get_return_type(LiteralNode.transform)


class RepeatTerm(Term):
    def __init__(
        self,
        child: Term,
        min_count: int,
        max_count: typing.Optional[int]
    ) -> None:
        self.child = child
        self.min_count = min_count
        self.max_count = max_count

    def __repr__(self) -> str:
        return '<RepeatTerm(Term) {0!r},{1},{2}>'.format(
            self.child,
            self.min_count,
            self.max_count
        )

    def match(
        self,
        text: _SmartText,
        lang: Language,
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        match, more_funcs = self._match_impl(
            text,
            lang,
            [],
            cache,
        )
        if match:
            yield match
        for func in more_funcs:
            match, even_more_funcs = func()
            if match:
                lang.log_match(match[0], 'repeat', len(match[1].children))
                yield match
            more_funcs += even_more_funcs

    def _match_impl(
        self,
        text: _SmartText,
        lang: Language,
        matched_terms: typing.Sequence[Node],
        cache: RuleReferenceCache,
    ) -> typing.Tuple[
        typing.Optional[typing.Tuple[_SmartText, MultiNode]],
        typing.Iterable[typing.Callable[[], typing.Any]]
    ]:
        if self.max_count is not None and len(matched_terms) > self.max_count:
            return None, []
        if len(matched_terms) >= self.min_count:
            match: typing.Optional[typing.Tuple[_SmartText, MultiNode]] = (
                text, MultiNode(matched_terms)
            )
        else:
            match = None

        more_funcs = [
            partial(
                self._match_impl,
                leftover_text,
                lang,
                list(matched_terms) + [term],
                cache,
            ) for leftover_text, term in self.child.match(text, lang, cache)
        ]
        return match, more_funcs

    def equals(self, other: Term) -> Validity:
        if not isinstance(other, RepeatTerm):
            return Validity.invalid('RepeatTerm: other is not a RepeatTerm')
        elif (
            self.min_count != other.min_count
            or self.max_count != other.max_count
        ):
            return Validity.invalid(
                'RepeatTerm: other does not have same range'
            )
        validity = self.child.equals(other.child)
        if not validity:
            return validity + Validity.invalid(
                'RepeatTerm: children are not equal'
            )
        else:
            return Validity.valid()

    def validate(self, lang: Language) -> Validity:
        if (
            self.max_count is not None
            and self.max_count < self.min_count
        ):
            return Validity.invalid(
                'RepeatTerm: max count is less than min count'
            )
        elif self.max_count is not None and self.max_count < 0:
            return Validity.invalid(
                'RepeatTerm: max count cannot be negative'
            )
        elif self.min_count < 0:
            return Validity.invalid(
                'RepeatTerm: min count cannot be negative'
            )
        validity = self.child.validate(lang)
        if not validity:
            return validity + Validity.invalid(
                'RepeatTerm: child is not valid'
            )
        else:
            return Validity.valid()

    def get_transform_type(self, lt: 'LanguageTransformation') -> type:
        return typing.Sequence[self.child.get_transform_type(lt)]  # type: ignore


class GroupTerm(Term):
    def __init__(
        self,
        children_groups: typing.Sequence[typing.Sequence[Term]]
    ) -> None:
        self.children_groups = children_groups

    def __repr__(self) -> str:
        return '<GroupTerm(Term) {0}>'.format(
            ','.join(
                '({0})'.format(
                    ','.join(repr(c) for c in children)
                ) for children in self.children_groups
            )
        )

    def equals(self, other: object) -> Validity:
        if not isinstance(other, GroupTerm):
            return Validity.invalid(
                'GroupTerm: other is not a GroupTerm'
            )
        elif len(self.children_groups) != len(other.children_groups):
            return Validity.invalid(
                'GroupTerm: this term has {0} children_groups and other has {1} '
                'children_groups'.format(
                    len(self.children_groups), len(other.children_groups)
                )
            )
        validity = Validity.valid()
        for index, (a, b) in enumerate(
            zip(self.children_groups, other.children_groups)
        ):
            if len(a) != len(b):
                validity += Validity.invalid(
                    'GroupTerm: children_groups {0} are not the same length'
                    .format(index)
                )
            else:
                validity += sum(
                    (c.equals(o) for c, o in zip(a, b)),
                    Validity.valid()
                )
        return validity

    def match(
        self,
        text: _SmartText,
        lang: Language,
        cache: RuleReferenceCache,
    ) -> typing.Iterable[MatchResult]:
        for index, children in enumerate(self.children_groups):
            for leftover, match in _group_match(children, text, lang, cache):
                yield leftover, MultiNode(match, (index, len(self.children_groups)))

    def validate(self, lang: Language) -> Validity:
        # TODO: validate if the grouping term is necessary, but requires extra context
        if len(self.children_groups) <= 1 and len(self.children_groups[0]) <= 1:
            return Validity.invalid(
                'GroupTerm: must have at least two children'
            )
        children = [
            c.validate(lang)
            for children in self.children_groups
            for c in children
        ]
        if all(children):
            return Validity.valid()
        else:
            msg_groups = [
                list(c.messages) + [
                    'GroupTerm[{0}]: invalid child'.format(index)
                ]
                for index, c in enumerate(children) if not c
            ]
            return (
                Validity(
                    [msg for group in msg_groups for msg in group]
                ) + Validity.invalid('GroupTerm: children are not valid')
            )

    def get_transform_type(self, lt: 'LanguageTransformation') -> type:  # pylint: disable=too-many-return-statements,too-many-branches
        child_types = tuple(
            typing.Tuple[tuple(
                c.get_transform_type(lt) for c in children
            )] for children in self.children_groups
        )
        # c0, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c13, c14, c15 = \
        #     itertools.islice(
        #         itertools.chain(child_types_gen, itertools.repeat(None))
        #         16
        #     )
        if len(self.children_groups) == 1:
            return gt.Group[child_types]  # type: ignore
        elif len(self.children_groups) == 2:
            return gt.Group2[child_types]  # type: ignore
        elif len(self.children_groups) == 3:
            return gt.Group3[child_types]  # type: ignore
        elif len(self.children_groups) == 4:
            return gt.Group4[child_types]  # type: ignore
        elif len(self.children_groups) == 5:
            return gt.Group5[child_types]  # type: ignore
        elif len(self.children_groups) == 6:
            return gt.Group6[child_types]  # type: ignore
        elif len(self.children_groups) == 7:
            return gt.Group7[child_types]  # type: ignore
        elif len(self.children_groups) == 8:
            return gt.Group8[child_types]  # type: ignore
        elif len(self.children_groups) == 9:
            return gt.Group9[child_types]  # type: ignore
        elif len(self.children_groups) == 10:
            return gt.Group10[child_types]  # type: ignore
        elif len(self.children_groups) == 11:
            return gt.Group11[child_types]  # type: ignore
        elif len(self.children_groups) == 12:
            return gt.Group12[child_types]  # type: ignore
        elif len(self.children_groups) == 13:
            return gt.Group13[child_types]  # type: ignore
        elif len(self.children_groups) == 14:
            return gt.Group14[child_types]  # type: ignore
        elif len(self.children_groups) == 15:
            return gt.Group15[child_types]  # type: ignore
        elif len(self.children_groups) == 16:
            return gt.Group16[child_types]  # type: ignore
        else:
            raise ValueError('cant have more than 16 child groups')


def _group_match(
    children: typing.Sequence[Term],
    text: _SmartText,
    lang: Language,
    cache: RuleReferenceCache,
) -> typing.Iterable[typing.Tuple[_SmartText, typing.Sequence[Node]]]:
    funcs: typing.List[typing.Callable[[], typing.Any]] = [
        partial(_group_match_impl, children, text, lang, [], cache)
    ]
    for f in funcs:
        result: typing.Union[
            typing.Tuple[_SmartText, typing.Sequence[Node]],
            typing.List[typing.Callable[[], typing.Any]]
        ] = f()
        if isinstance(result, tuple):
            yield result
        else:
            funcs += result


def _group_match_impl(
    children: typing.Sequence[Term],
    text: _SmartText,
    lang: Language,
    matched_terms: typing.Sequence[Node],
    cache: RuleReferenceCache,
) -> typing.Union[
    typing.Tuple[_SmartText, typing.Sequence[Node]],
    typing.List[typing.Callable[[], typing.Any]]
]:
    if len(matched_terms) == len(children):
        return text, matched_terms
    else:
        target = children[len(matched_terms)]
        return [
            partial(
                _group_match_impl,
                children,
                leftover,
                lang,
                list(matched_terms) + [node],
                cache,
            ) for leftover, node in target.match(text, lang, cache)
        ]



import typing
from typing import Tuple

from ...core import grammar as g, transform as t
from ...validation.transform_validation import annotate
from ...validation import group_types as gt, switches as sw

from ._transform_helpers import (
    nothing, nothing2, nothing3, identity, identity2, add, unescape)
from ._transform_terminals import add_terminal_transforms
from ._freebies import add_freebie_rules, add_freebie_transforms


def rule_reference_accum(values: typing.Tuple[str]) -> g.Term:
    return g.RuleReference(values[0])


def syntax_accum(
    values: typing.Tuple[
        typing.Sequence[
            typing.Union[g.Rule, typing.Tuple[str, typing.List[g.TermGroup]]]
        ]
    ]
) -> g.Language:
    rules = values[0]
    if not isinstance(rules[0], g.Rule):
        raise TypeError('First rule cannot use "=/"')
    lang = g.Language.create(rules[0].name)
    lang = add_freebie_rules(lang)
    for rule in rules:
        if isinstance(rule, g.Rule):
            lang.add_rule(rule)
        else:
            lang.add_to_rule(rule[0], rule[1])
    return lang


def rule_accum(
    values: typing.Tuple[
        None,
        str,
        None,
        str,
        None,
        typing.List[g.TermGroup],
        typing.Sequence[None]
    ]
) -> typing.Union[g.Rule, typing.Tuple[str, typing.List[g.TermGroup]]]:
    if values[3] == '=':
        return g.Rule(values[1], g.Syntax(values[5]))
    elif values[3] == '=/':
        return values[1], values[5]
    else:
        raise RuntimeError


def expression_accum(
    values: typing.Tuple[
        g.TermGroup,
        typing.Sequence[g.TermGroup]
    ]
) -> typing.List[g.TermGroup]:
    # TO DO: return a list of lists instead
    return [values[0]] + list(values[1])


def list_accum(
    values: typing.Tuple[g.Term, typing.Sequence[g.Term]]
) -> g.TermGroup:
    return g.TermGroup([values[0]] + list(values[1]))


def text_accum_1(
    values: typing.Tuple[str]
) -> g.Literal:
    return g.Literal(values[0])


def text_accum_2(
    values: typing.Tuple[str, g.Literal]
) -> g.Literal:
    return g.Literal(values[0] + values[1].text)


def literal_range_accum1(
    values: typing.Tuple[str, str, str]
) -> g.LiteralRange:
    number = int(values[1])
    return g.LiteralRange(number, number)


def literal_range_accum2(
    values: typing.Tuple[str, str, str, str, str]
) -> g.LiteralRange:
    first = int(values[1])
    second = int(values[3])
    return g.LiteralRange(first, second)


def repeat_body_accum1(
    values: typing.Tuple[typing.Sequence[str], str, typing.Sequence[str]]
) -> typing.Tuple[int, typing.Optional[int]]:
    if values[0]:
        start = int(values[0][0])
    else:
        start = 0
    if values[2]:
        end: typing.Optional[int] = int(values[2][0])
    else:
        end = None
    return (start, end)


def repeat_body_accum2(
    values: typing.Tuple[str]
) -> typing.Tuple[int, typing.Optional[int]]:
    num = int(values[0])
    return (num, num)


def expression_end_accum(
    values: typing.Tuple[None, str, None, g.TermGroup]
) -> g.TermGroup:
    return values[3]


def list_end_accum(
    values: typing.Tuple[typing.Sequence[str], g.Term]
) -> g.Term:
    return values[1]


def number_accum(
    values: typing.Tuple[str, typing.Sequence[str]]
) -> str:
    val = values[0]
    for x in values[1]:
        val += x
    return val


def rule_name_accum(
    values: typing.Tuple[str, typing.Sequence[str]]
) -> str:
    val = values[0]
    for x in values[1]:
        val += x
    return val


def text_accum(
    values: typing.Tuple[typing.Sequence[str]]
) -> g.Literal:
    iterable = iter(values[0])
    accum = next(iterable)
    for item in iterable:
        accum += item
    return g.Literal(accum)


def _repeat_term_accum1(
    values: typing.Tuple[typing.Tuple[int, typing.Optional[int]], g.Term]
) -> g.Term:
    ((start, end), term) = values
    return g.RepeatTerm(term, start, end)


def _repeat_term_accum2(
    values: typing.Tuple[str, g.Term, str]
) -> g.Term:
    return g.RepeatTerm(values[1], 0, 1)


class _StringLiteralSwitch(sw.Switch2[Tuple[str], Tuple[str], bool]):
    @staticmethod
    def case0(val: Tuple[str]) -> bool:
        assert val[0] == '%s'
        return True

    @staticmethod
    def case1(val: Tuple[str]) -> bool:
        assert val[0] == '%i'
        return False


def _string_literal_accum(
    values: typing.Tuple[
        typing.Sequence[
            gt.Group2[
                typing.Tuple[str],
                typing.Tuple[str]
            ]
        ],
        typing.Sequence[str]
    ]
) -> g.Literal:
    value = ''.join(values[1])
    case_sensitive = False
    if values[0]:
        case_sensitive = _StringLiteralSwitch()(values[0][0])
    return g.Literal(value, case_sensitive)


def _string_literal_accum2(
    values: typing.Tuple[str, typing.Sequence[str]]
) -> g.Literal:
    (sensitivity, list_value) = values
    value = ''.join(list_value)
    if sensitivity == "%s":
        return g.Literal(value, True)
    elif sensitivity == "%i":
        return g.Literal(value, False)
    else:
        raise RuntimeError('unreachable')


def _group_term_accum(
    values: typing.Tuple[str, None, typing.List[g.TermGroup], None, str]
) -> g.Term:
    return g.GroupTerm([
        term_group.terms
        for term_group in values[2]
    ])


transform = t.LanguageTransformation.create('Syntax', [syntax_accum])
transform <<= 'Rule', [rule_accum]
transform <<= 'OptWhitespace', [
    annotate(nothing, T=typing.Sequence[str])
]
transform <<= 'Expression', [
    expression_accum
]
transform <<= 'SingleLineEnd', [
    annotate(nothing3, T=None, T2=typing.Sequence[None], T3=str)
]
transform <<= 'List', [
    list_accum
]
transform <<= 'RepeatableTerm', [
    annotate(identity2, T=g.Literal, T2=g.Term),
    annotate(identity, T=g.Term),
    annotate(identity, T=g.Term),
    annotate(identity, T=g.Term),
    rule_reference_accum,
    annotate(identity, T=g.Term),
]
transform <<= 'LiteralBody', [
    annotate(unescape, T=typing.Sequence[str])
]
transform <<= 'Character', [
    annotate(identity, T=str)
] * 3
transform <<= 'Letter', [
    annotate(identity, T=str)
] * 2
transform <<= 'Digit', [
    annotate(identity, T=str)
] * 2
transform <<= 'NonZeroDigit', [
    annotate(identity, T=str)
] * 1
transform <<= 'Symbol', [
    annotate(identity, T=str)
] * 6
transform <<= 'RuleName', [
    rule_name_accum
]
transform <<= 'OneRuleEnd', [
    annotate(identity, T=str),
    annotate(identity, T=str),
    annotate(add, Addable=str),
    annotate(add, Addable=str)
]
transform <<= 'Number', [
    annotate(identity, T=str),
    number_accum
]
transform <<= 'Term', [
    annotate(identity, T=g.Term),
    annotate(identity, T=g.Term),
]
transform <<= 'RepeatBody', [
    repeat_body_accum1,
    repeat_body_accum2,
]
transform <<= 'ExpressionEnd', [
    expression_end_accum
]
transform <<= 'ListEnd', [
    list_end_accum
]
transform <<= 'RepeatTerm', [
    _repeat_term_accum1,
    _repeat_term_accum2
]
transform <<= 'AssignmentOperator', [annotate(identity, T=str)] * 2
transform <<= 'Comment', [
    annotate(
        nothing2,
        T=str,
        T2=typing.Sequence[
            gt.Group2[typing.Tuple[str], typing.Tuple[str]]
        ]
    )
]
transform <<= 'StringLiteral', [
    _string_literal_accum,
]
transform <<= 'GroupTerm', [_group_term_accum]
transform = add_terminal_transforms(transform)
transform = add_freebie_transforms(transform)



import abc
from collections import deque
import functools
import typing

InputType = typing.TypeVar('InputType')
OutputType = typing.TypeVar('OutputType')


class Resolvable(typing.Generic[OutputType], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def resolve(self, d: deque) -> typing.Union['Resolvable[OutputType]', OutputType]:
        pass


class ResolvableFunc(functools.partial, Resolvable[OutputType]):
    def resolve(self, _: deque) -> typing.Union[Resolvable[OutputType], OutputType]:
        return self()


class ResolvablePair(Resolvable[OutputType], typing.Generic[InputType, OutputType]):
    def __init__(
        self,
        resolvable: Resolvable[InputType],
        next_resolvable: typing.Callable[[InputType], Resolvable[OutputType]]
    ) -> None:
        self.resolvable = resolvable
        self.next_resolvable = next_resolvable

    def resolve(self, d: deque) -> typing.Union[Resolvable[OutputType], OutputType]:
        d.append(self.next_resolvable)
        return self.resolvable  # type: ignore


def _identity(x: OutputType) -> OutputType:
    return x


def resolve(result: typing.Any) -> OutputType:
    d = deque([_identity])
    while d:
        result = d.pop()(result)
        while isinstance(result, Resolvable):
            result = result.resolve(d)  # typing: ignore
    return result


def resolve_map(
    values: typing.Sequence[InputType],
    func: typing.Callable[[InputType], OutputType]
) -> Resolvable[typing.List[OutputType]]:
    return _resolve_map_impl(list(values), func, 0, [])


def _resolve_map_impl(
    values: typing.Sequence[InputType],
    func: typing.Callable[[InputType], OutputType],
    index: int,
    result: typing.List[OutputType]
) -> Resolvable[typing.List[OutputType]]:
    if index >= len(values):
        return ResolvableFunc(
            _identity,
            result
        )
    else:
        return ResolvablePair(
            ResolvableFunc(func, values[index]),
            functools.partial(
                _update_result,
                values,
                func,
                index,
                result
            )
        )


def _update_result(
    values: typing.Sequence[InputType],
    func: typing.Callable[[InputType], OutputType],
    index: int,
    result: typing.List[OutputType],
    value: OutputType,
) -> Resolvable[typing.List[OutputType]]:
    return ResolvableFunc(
        _resolve_map_impl,
        values,
        func,
        index + 1,
        result + [value]
    )



import abc
import re
import typing

from .resolvable import resolve_map, ResolvablePair, ResolvableFunc, Resolvable
from ..validation import group_types as gt

if typing.TYPE_CHECKING:
    from .transform import LanguageTransformation  # pylint: disable=unused-import

RE_LINE_START = re.compile('^', flags=re.MULTILINE)
RuleName = str


class Node(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def transform(self, lang: 'LanguageTransformation') -> typing.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def draw(self) -> str:
        raise NotImplementedError


class RuleNode(Node):
    def __init__(
        self,
        matched_rule: RuleName,
        term_group_id: int,
        children: typing.Sequence[Node]
    ) -> None:
        self.matched_rule = matched_rule
        self.term_group_id = term_group_id
        self.children = children

    def __str__(self) -> str:
        return ''.join(str(c) for c in self.children)

    def __repr__(self) -> str:
        return '<RuleNode {0}>'.format(repr(self.matched_rule))

    def transform(self, lang: 'LanguageTransformation') -> typing.Any:
        return lang.transformation_rules[self.matched_rule].transform(
            self,
            lang
        )

    def draw(self) -> str:
        children = (
            RE_LINE_START.sub('  ', child.draw()) for child in self.children
        )
        return '\n'.join(
            [
                repr(self),
                *children
            ]
        )


class LiteralNode(Node):
    def __init__(self, value: str) -> None:
        self.value = value

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return '<LiteralNode {0}>'.format(repr(self.value))

    def transform(self, lang: 'LanguageTransformation') -> str:
        return self.value

    def draw(self) -> str:
        return repr(self)


class MultiNode(Node):
    def __init__(
        self,
        children: typing.Sequence[Node],
        group_info: typing.Optional[typing.Tuple[int, int]] = None,
    ) -> None:
        self.children = children
        self.group_info = group_info
        # group_info[0] index of the group that matched
        # group_info[1] total number of groups

    def transform(self, lang: 'LanguageTransformation') -> typing.Any:
        resolvable_result = resolve_map(
            self.children,
            lambda c: c.transform(lang)
        )
        if self.group_info is None:
            return resolvable_result
        else:
            return self._make_group_result(
                resolvable_result
            )

    def __str__(self) -> str:
        return ''.join(str(child) for child in self.children)

    def __repr__(self) -> str:
        if not self.children:
            return '<MultiNode 0>'
        else:
            return '<MultiNode {0}>'.format(
                ','.join(repr(c) for c in self.children)
            )

    def draw(self) -> str:
        children = (
            RE_LINE_START.sub('  ', child.draw()) for child in self.children
        )
        return '\n'.join(
            [
                repr(self),
                *children
            ]
        )

    def _make_group_result(self, resolvable: Resolvable) -> Resolvable:
        def func(result: typing.Any) -> Resolvable:
            def nested() -> typing.Any:
                if self.group_info:
                    output = [gt.NoValue.Sentinel] * self.group_info[1]
                    output[self.group_info[0]] = result
                    return output
                else:
                    raise RuntimeError

            return ResolvableFunc(nested)

        return ResolvablePair(resolvable, func)



from prosodia.core.grammar import Grammar, Language

from .. import create_bnfrepeat
from ._text import example_text
from ._transform import transform


def create_example_bnfrepeat() -> Grammar[Language]:
    return Grammar(create_bnfrepeat().apply(example_text), transform)



#!/usr/bin/env python3

from setuptools import setup, find_packages

test_requirements = [
    'pytest-cov',
    'pytest'
]

setup(
    name='prosodia',
    version='0.5.3',
    author='macbeth322',
    author_email='chrisp533@gmail.com',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=[
        'typing'
    ],
    setup_requires=['pytest-runner'],
    tests_require=test_requirements,
    extras_require={
        'test': test_requirements
    },
    package_data={
        'prosodia': ['py.typed'],
        '': ['*.grammar']
    },
    zip_safe=False
)



from typing import Tuple, Sequence, Union, TYPE_CHECKING, TypeVar, Callable

from ...core import grammar as g, transform as t
from ...validation.transform_validation import annotate

from ._transform_helpers import identity, identity2

if TYPE_CHECKING:
    T = TypeVar('T')


def _terminal_accum(
    values: Tuple[str, int, Sequence[
        Union[int, Sequence[int]]
    ]]
) -> g.Term:
    _, first, optional_tail = values
    if optional_tail:
        (tail,) = optional_tail
        if isinstance(tail, int):
            return g.LiteralRange(first, tail)
        else:
            return g.Literal(''.join(
                chr(x) for x in [first] + list(tail)
            ))
    else:
        return g.Literal(chr(first))


def _partial_str_to_int(base: int) -> Callable[[Tuple[Sequence[str]]], int]:
    def func(values: Tuple[Sequence[str]]) -> int:
        return int(''.join(values[0]), base=base)

    return func


def _tail_range_accum(values: Tuple[str, int]) -> Union[int, Sequence[int]]:
    return values[1]


def _concat_unit_accum(values: Tuple[str, int]) -> int:
    return values[1]


def add_terminal_transforms(
    lt: t.LanguageTransformation
) -> t.LanguageTransformation:
    lt <<= 'BinaryLiteral', [_terminal_accum]
    lt <<= 'BinaryBody', [_partial_str_to_int(2)]
    lt <<= 'BinaryTail', [
        _tail_range_accum,
        annotate(identity2, T=Sequence[int], T2=Union[int, Sequence[int]])
    ]
    lt <<= 'BinaryConcatUnit', [_concat_unit_accum]
    lt <<= 'HexadecimalLiteral', [_terminal_accum]
    lt <<= 'HexadecimalBody', [_partial_str_to_int(16)]
    lt <<= 'HexadecimalTail', [
        _tail_range_accum,
        annotate(identity2, T=Sequence[int], T2=Union[int, Sequence[int]])
    ]
    lt <<= 'HexadecimalBodyUnit', [
        annotate(identity, T=str),
        annotate(identity, T=str)
    ]
    lt <<= 'HexadecimalConcatUnit', [_concat_unit_accum]
    lt <<= 'DecimalLiteral', [_terminal_accum]
    lt <<= 'DecimalBody', [_partial_str_to_int(10)]
    lt <<= 'DecimalTail', [
        _tail_range_accum,
        annotate(identity2, T=Sequence[int], T2=Union[int, Sequence[int]])
    ]
    lt <<= 'DecimalConcatUnit', [_concat_unit_accum]
    return lt



import typing

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')
    T2 = typing.TypeVar('T2')
    T3 = typing.TypeVar('T3')
    Addable = typing.TypeVar('Addable', str, typing.List[object])


def list_of(
    values: typing.Tuple['T']
) -> typing.List['T']:
    return [values[0]]


def push_list(
    values: typing.Tuple['T', typing.List['T']]
) -> typing.List['T']:
    return [values[0]] + values[1]


def nothing(_: typing.Tuple['T']) -> None:
    return None


def nothing2(_: typing.Tuple['T', 'T2']) -> None:
    return None


def nothing3(_: typing.Tuple['T', 'T2', 'T3']) -> None:
    return None


def identity(values: typing.Tuple['T']) -> 'T':
    return values[0]


def identity2(values: typing.Tuple['T']) -> 'T2':
    return values[0]  # type: ignore


def unescape(
    values: typing.Tuple[str, 'T', str]
) -> 'T':
    return values[1]


def add(
    values: typing.Tuple['Addable', 'Addable']
) -> 'Addable':
    iterable = iter(values)
    accum = next(iterable)
    for item in iterable:
        accum += item
    return accum



from prosodia.core.grammar import Grammar, Language

from ._text import text
from ._transform import transform
from ._intermediate_parser import create_intermediate_augmentedbnf
from ._freebies import add_freebie_rules


def create_language() -> Language:
    lang = create_intermediate_augmentedbnf().apply(text)
    add_freebie_rules(lang)
    return lang


def create_augmentedbnf() -> Grammar[Language]:
    return Grammar(create_language(), transform, False)



from enum import Enum
from typing import TypeVar, Union, Tuple


T0 = TypeVar('T0')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')
T6 = TypeVar('T6')
T7 = TypeVar('T7')
T8 = TypeVar('T8')
T9 = TypeVar('T9')
T10 = TypeVar('T10')
T11 = TypeVar('T11')
T12 = TypeVar('T12')
T13 = TypeVar('T13')
T14 = TypeVar('T14')
T15 = TypeVar('T15')


class NoValue(Enum):
    Sentinel = 0


UNV = Union[NoValue, T0]

Group = Tuple[T0]
Group2 = Tuple[UNV[T0], UNV[T1]]
Group3 = Tuple[UNV[T0], UNV[T1], UNV[T2]]
Group4 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3]]
Group5 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4]]
Group6 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5]]
Group7 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6]]
Group8 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
               UNV[T7]]
Group9 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
               UNV[T7], UNV[T8]]
Group10 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
                UNV[T7], UNV[T8], UNV[T9]]
Group11 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
                UNV[T7], UNV[T8], UNV[T9], UNV[T10]]
Group12 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
                UNV[T7], UNV[T8], UNV[T9], UNV[T10], UNV[T11]]
Group13 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
                UNV[T7], UNV[T8], UNV[T9], UNV[T10], UNV[T11], UNV[T12]]
Group14 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
                UNV[T7], UNV[T8], UNV[T9], UNV[T10], UNV[T11], UNV[T12],
                UNV[T13]]
Group15 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
                UNV[T7], UNV[T8], UNV[T9], UNV[T10], UNV[T11], UNV[T12],
                UNV[T13], UNV[T14]]
Group16 = Tuple[UNV[T0], UNV[T1], UNV[T2], UNV[T3], UNV[T4], UNV[T5], UNV[T6],
                UNV[T7], UNV[T8], UNV[T9], UNV[T10], UNV[T11], UNV[T12],
                UNV[T13], UNV[T14], UNV[T15]]
